/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 *
 * dos2unix '\n' convertor 0.5.0
 *   based on Unix2Dos 0.9.0 by Peter Hanecak (made 19.2.1997)
 * Copyright 1997,.. by Peter Hanecak <hanecak@megaloman.sk>.
 * All rights reserved.
 *
 * Some portions by Rob Ginda <rginda@netscape.com> (07.2.2001)
 *
 * dos2unix filters reading input from stdin and writing output to stdout.
 * Without arguments it reverts the format (e.i. if source is in UNIX format,
 * output is in DOS format and vice versa).
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * See the COPYING file for license information.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>

#define CT_AUTO 0
#define CT_UNIX2DOS 1
#define CT_DOS2UNIX 2

#define FT_MIXED 0
#define FT_DOS   1
#define FT_UNIX  2

static struct option long_options[] = {
    { "auto", no_argument, NULL, 'A' },
    { "u2d", no_argument, NULL, 'D' },
    { "d2u", no_argument, NULL, 'U' },
    { "help", no_argument, NULL, 'h' },
    { "verbose", no_argument, NULL, 'v' },
    { "test", no_argument, NULL, 't' },
    { 0, 0, 0, 0 }
};

char testmode = 0, verbose = 0;
char *argv0;

/* if fn is NULL then input is stdin and output is stdout */
int convert(char *fn, int convType) {
    char c, fileType;
    char *tempFn = NULL;
    uint crlfCount = 0, lfCount = 0;
    FILE *in = stdin, *out = stdout;

    /* ugly little macro to clean up and return */
#define BAIL(_rval, _msg) {                                                    \
        if (in && in != stdin)                                                 \
            if (fclose(in) < 0) {                                              \
                perror(argv0); return -2;                                      \
            }                                                                  \
        if (out && out != stdout)                                              \
            if (fclose(out) < 0) {                                             \
                perror(argv0); return -2;                                      \
            }                                                                  \
        if (tempFn && remove(tempFn) < 0) {                                    \
            perror(argv0); return -2;                                          \
        }                                                                      \
        if (_rval != 0)                                                        \
            perror(_msg);                                                      \
        return _rval;                                                          \
    }

    /* open the input file, if it isn't stdin */
    if (fn != NULL &&
        (in = fopen(fn, "r")) == NULL)
        BAIL (-1, fn);

    /* initial scan for filetype */
    while ((c = fgetc(in)) != EOF)
        if (c == '\r') {
            if (fgetc(in) != '\n') {
                if (verbose)
                    fprintf (stderr, "Rogue \\r in input file %s, skipping.\n",
                             fn);
                BAIL (0, NULL);
            }
            ++crlfCount;
        } else if (c == '\n') {
            ++lfCount;
        } else if ((c < 32 || c > 126) && c != '\t') {
            if (verbose)
                fprintf (stderr,
                         "Input file '%s' looks binary (%i), skipping.\n",
                         fn, (int)c);
            else if (testmode)
                fprintf (stderr, "Input file '%s' looks binary (%i).\n",
                         fn, (int)c);
            BAIL (0, NULL);
        }

    if (crlfCount == 0 && lfCount == 0) {
            if (verbose)
                fprintf (stderr,
                         "Input file '%s' has no line endings, skipping.\n",
                         fn);
            else if (testmode)
                fprintf (stderr,
                         "Input file '%s' has no line endings.\n", fn);
            BAIL (0, NULL);
    }

    /* figure out the source file type (mixed means both kinds of line endings
     * were found.) */
    fileType = (crlfCount != 0 && lfCount != 0) ? FT_MIXED :
        (crlfCount > lfCount) ? FT_DOS : FT_UNIX;

    /* if we're autodetecting, and the source has dos or mixed endings,
     * do a dos -> unix conversion, otherwise unix -> dos */
    if (convType == CT_AUTO)
        convType = (fileType == FT_MIXED || fileType == FT_DOS) ?
            CT_DOS2UNIX : CT_UNIX2DOS;

    /* don't waste time on a null conversion */
    if ((fileType == FT_DOS && convType == CT_UNIX2DOS) ||
        (fileType == FT_UNIX && convType == CT_DOS2UNIX))
    {
        if (verbose)
            fprintf (stderr,
                     "Input file '%s' is already in %s format, skipping.\n",
                     fn, (fileType == FT_DOS) ? "dos" : "unix");
        BAIL (0, NULL);
    }

    if (testmode) {
        fprintf (stderr, "Input file '%s' currently uses %s line endings.\n",
                 fn, (fileType == FT_MIXED) ? "mixed" :
                 (fileType == FT_DOS) ? "dos" : "unix",
                 (convType == CT_UNIX2DOS) ? "dos" : "unix");
        BAIL (0, NULL);
    }

    if (verbose)
        fprintf (stderr, "Converting '%s' from %s to %s line endings.\n", fn,
                 (fileType == FT_MIXED) ? "mixed" :
                 (fileType == FT_DOS) ? "dos" : "unix",
                 (convType == CT_UNIX2DOS) ? "dos" : "unix");

    /* if input isn't stdin, seek back to the beginning for the second pass,
     * and open a temp file for the output */
    if (fn != NULL) {
        if (fseek (in, 0, SEEK_SET) != 0)
            BAIL (-3, argv0);

        tempFn = tmpnam(NULL);
        if ((out = fopen(tempFn, "w")) == NULL)
            BAIL (-2, argv0);
    }

    /* do the actual conversion */
    while ((c = fgetc(in)) != EOF) {
        if (convType == CT_DOS2UNIX && c == '\r')
            continue;
        else if (convType == CT_UNIX2DOS && c == '\n')
            fputc('\r', out);
        fputc(c, out);
    }

    /* if we're not working with stdin/out, close source (in) and temp file
     * (out), and copy the temp file over the source. */
    if (fn != NULL) {
        if (fclose(in) != 0, in = 0)
            BAIL (-2, argv0);
        if (fclose(out) != 0, out = 0)
            BAIL (-2, argv0);

        if ((in = fopen(tempFn, "r")) == NULL)
            BAIL (-2, argv0);
        if ((out = fopen(fn, "w")) == NULL)
            BAIL (-2, argv0);

        while ((c = fgetc(in)) != EOF)
            fputc(c, out);
    }

    BAIL (0, NULL);

#undef BAIL

}

void help() {
    fprintf(stderr,
            "usage:\t%s [--verbose|-v] [--test|-t] [--d2u|--u2d|--auto|-D|-U|-A]\n"
            "\t\t [<file name> [...]]\n"
            "\t\t--auto, -A\toutput will be set based upon autodetection\n"
            "\t\t\t\tof source format\n"
            "\t\t--d2u, -U\toutput will be in UNIX format\n"
            "\t\t--u2d, -D\toutput will be in DOS format\n"
            "\t\t--test, -t\tdon't actually do a conversion,\n"
            "\t\t\t\tjust report on source type\n"
            "\t\t--verbose, -v\tprint extra information on stderr\n"
            "\n"
            "- when no options are given then input format will be automaticaly detected\n"
            "  and converted to the opposite format on output\n"
            "- when no file is given, then stdin is used as input and"
            "  stdout as output\n"
            "- binary files will be skipped automatically\n", argv0);
}

int main(int argc, char *argv[]) {
    int convType = CT_AUTO;
    int o;

    argv0 = argv[0];

    /* process parameters */
    while ((o = getopt_long(argc, argv, "ADUhvt", long_options, NULL)) != EOF) {
        switch (o) {
            case 'A':
                convType = CT_AUTO;
                break;
            case 'D':
                convType = CT_UNIX2DOS;
                break;
            case 'U':
                convType = CT_DOS2UNIX;
                break;
            case 'h':
                help();
                return 0;
            case 'v':
                verbose = 1;
                break;
            case 't':
                testmode = 1;
                break;
            case '?':
                help();
                return -1;
        }
    }

    if (optind < argc) {
        while(optind < argc)
            if ((o = convert(argv[optind++], convType)) != 0)
                break;
    }
    else
        o = convert(NULL, convType);

    return o;
}
